# Michael Halle
# mhalle@bwh.harvard.edu
# version 0.3.0

__doc__="""DTML Template objects."""

import OFS.ObjectManager
from DocumentTemplate.DT_Var import Var, Comment
from OFS.content_types import guess_content_type
from DocumentTemplate.DT_Util import ParseError, parse_params, name_param
from Globals import HTML, HTMLFile, MessageDialog
from OFS.DTMLMethod import DTMLMethod, decapitate
from DocumentTemplate.DT_HTML import dtml_re_class
from urllib import quote
import DT_Block
from DT_Block import Block
import Globals
import sys

TEMPLATE_BLOCK_TYPE='DTML Template block'
TEMPLATE_TYPE='DTML Template'

class DTMLTemplate(OFS.ObjectManager.ObjectManager,
                   DTMLMethod):
    """DTML Template objects are TemplateTemplate.HTML objects that act
       as methods whose 'self' is the DTML Template itself."""

    meta_type=TEMPLATE_TYPE

    manage_options=({'label':'Edit', 'action':'manage_main'},
                    {'label':'Upload', 'action':'manage_uploadForm'},
                    {'label':'View', 'action':''},
                    {'label':'Proxy', 'action':'manage_proxyForm'},
                    {'label':'Security', 'action':'manage_access'},
                   )

    __ac_permissions__=(
        ('Change DTML Templates',   ('manage_edit', 'manage_upload', 'PUT')),
    )

    manage_editForm=HTMLFile('templateEdit', globals())
    manage_uploadForm=HTMLFile('templateUpload', globals())
    manage=manage_main=manage_editDocument=manage_editForm
    manage_proxyForm=HTMLFile('templateProxy', globals())

    def manage_edit(self,data,title,SUBMIT='Change',dtpref_cols='50',
                    dtpref_rows='20',REQUEST=None, asDocument=0):
        """
        Replaces a Template's contents with Data, Title with Title.

        The SUBMIT parameter is also used to change the size of the editing
        area on the default Template edit screen.  If the value is "Smaller",
        the rows and columns decrease by 5.  If the value is "Bigger", the
        rows and columns increase by 5.  If any other or no value is supplied,
        the data gets checked for DTML errors and is saved.
        """
        self._validateProxy(REQUEST)
        if self._size_changes.has_key(SUBMIT):
            return self._er(data,title,SUBMIT,dtpref_cols,dtpref_rows,REQUEST)
        self.title=title
        self._asDocument = asDocument
        if type(data) is not type(''): data=data.read()
        self.munge(data)
        self._findTemplateBlocks()
        if REQUEST:
            message="Content changed."
            return self.manage_main(self,REQUEST,manage_tabs_message=message)


    def index_html(self, REQUEST={}, RESPONSE=None, **kw):
        """Render the document given a client object, REQUEST mapping,
        Response, and key word arguments."""
        return apply(self.__call__, (self, REQUEST, RESPONSE), kw)
         

    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render the document given a client object, REQUEST mapping,
        Response, and key word arguments."""

        if hasattr(self, '_asDocument') and self._asDocument:
            # document behavior
            if hasattr(self, 'aq_explicit'): bself=self.aq_explicit
            else: bself=self

            if client is None:
                modified_client = bself
            else:
                modified_client = (client, bself)
        else:
            # method behavior
            modified_client = client

        return apply(DTMLMethod.__call__,
                     (self, modified_client, REQUEST, RESPONSE), kw)


    def manage_upload(self,file='', REQUEST=None):
        """Replace the contents of the template with the text in file."""
        self._validateProxy(REQUEST)
        if type(file) is not type(''): file=file.read()
        self.munge(file)
        self._findTemplateBlocks()
        if REQUEST: return MessageDialog(
                    title  ='Success!',
                    message='Your changes have been saved',
                    action ='manage_main')

    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests."""
        self.dav__init(REQUEST, RESPONSE)
        body=REQUEST.get('BODY', '')
        self._validateProxy(REQUEST)
        self.munge(body)
        self._findTemplateBlocks()        
        RESPONSE.setStatus(204)
        return RESPONSE

    def asDocument(self):
        """
        If true, template is using document semantics: methods
        are evaluated in the content of the template itself.
        """
        return (hasattr(self, '_asDocument') and self._asDocument)
        
    class namespace_: pass

    def blockIds(self):
        """Return a list of blocks in the template"""
        return self.objectIds(TEMPLATE_BLOCK_TYPE)

    def blocks(self, block_spec=None, **kw):
        """Returns a namespace that includes the blocks named in spec,
        which can be an id or a list of ids, or all blocks if spec
        isn't given."""
        
        r = self.namespace_()
        d = r.__dict__
        if block_spec is None and not kw:
            # return all blocks
            blocklist = self.objectItems(TEMPLATE_BLOCK_TYPE)
            for k, v in blocklist:
                d[k] = v
                
            return r
        
        if block_spec is not None:
            if type(block_spec) is type(''):
                block_spec = [block_spec]
            for k in block_spec:
                v = self._getOb(k)
                if v.meta_type is not TEMPLATE_BLOCK_TYPE:
                    raise KeyError, k
                d[k] = v

        if kw:
            for dk in kw.keys():
                k = kw[dk]
                v = self._getOb(k)
                if v.meta_type is not TEMPLATE_BLOCK_TYPE:
                    raise KeyError, k
                d[dk] = v

        return r

            
    def _findTemplateBlocks(self):
        for item in self.objectIds(TEMPLATE_BLOCK_TYPE):
            self._delObject(item)

        text = self.raw
        if text is None:
            return

        for bd in self.tmpl_parse(text):
            name = bd[0]
            block_text = text[bd[1]:(bd[2])]
            try:
                ob = TemplateBlock(name, file='', __name__=bd[0])
                self._setObject(bd[0], ob)
            except:
                raise 'Template Error', (
                    'the block name "%s" is not unique' % name)
            ob.manage_upload(block_text)


    def tmpl_parse(self,text,start=0,result=None,tagre=None,inComment=0):

        # this parsing function is based on the DTML parser,
        # but collects indicies of the beginning and end of
        # dtml-block blocks.  It uses the standard
        # DTML "parse_close" method since it does not change
        # any data structures.

        # The reason I reparse is because this way I dont
        # have to change the existing DTML parser.
        
        if tagre is None: tagre = dtml_re_class()
        if result is None: result = []
        l = tagre.search(text, start)

        while l >= 0:
            try: tag, args, command, coname = self._parseTag(tagre)
            except ParseError, m: self.parse_error(m[0],m[1],text,l)
            start=l+len(tag)
            if hasattr(command,'blockContinuations'):
                # we only care about blocks of text

                start=self.skip_eol(text,start)

                if command is Comment:
                    inComment = inComment + 1
                
                (pre,end)=self.tmpl_parse_block(text, start, result, tagre,
                                          tag, l, args, command, inComment)
                
                if command is Block and not inComment:
                    argd = DT_Block.parse_params(args)
                    if not argd.has_key('private'):
                        name, expr = DT_Block.name_param(argd)
                        result.append((name, start, pre))
                        
                if command is Comment:
                    inComment = inComment - 1

                start = end
            l=tagre.search(text,start)

        text=text[start:]
        return result

    def tmpl_parse_block(self, text, start, result, tagre,
                    stag, sloc, sargs, scommand, inComment):
        tname=scommand.name
        sname=stag
        sstart=start
        sa=sargs
        while 1:
            l=tagre.search(text,start)
            if l < 0: self.parse_error('No closing tag', stag, text, sloc)

            try: tag, args, command, coname = self._parseTag(tagre,scommand,sa)
            except ParseError, m: self.parse_error(m[0],m[1], text, l)
            if command:
                start=l+len(tag)
                if hasattr(command, 'blockContinuations'):
                    # New open tag.  Need to find closing tag.
                    start=self.parse_close(text, start, tagre, tag, l,
                                           command, args)
            else:
                # Either a continuation tag or an end tag
                result.extend(self.tmpl_parse(text[:l],sstart,
                                              inComment=inComment))

                start=self.skip_eol(text,l+len(tag))

                if coname:
                    tname=coname
                    sname=tag
                    sargs=args
                    sstart=start
                else:
                    return (l, start)

default_dd_html="""<dtml-var standard_html_header>
<h2><dtml-var title_or_id></h2>
<p>
This is the <dtml-var id> Template.
</p>
<dtml-var standard_html_footer>"""

addDTMLTemplateForm=HTMLFile('templateAdd', globals())

def addDTMLTemplate(self, id, title='', file='', REQUEST=None,
                    submit=None, asDocument=0):
    """Add a DTML Template object with the contents of file. If
    'file' is empty, default template text is used.
    """
    if type(file) is not type(''): file=file.read()
    if not file: file=default_dd_html
    ob=DTMLTemplate(file, __name__=id)
    ob.title=title
    ob._asDocument = asDocument
    
    id=self._setObject(id, ob)
    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u,quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_main')
    return ''


class TemplateBlock(DTMLMethod):
    "Trivial Subclass of DTMLMethod.  One per block in a DTML Template "
    meta_type=TEMPLATE_BLOCK_TYPE

    manage_options=({'label':'View', 'action':''},
                    )

    isPrincipiaFolderish = 0

    def _redirect_up(self, REQUEST, RESPONSE):
        """ pass of management to parent """

        if REQUEST is not None:
            u=REQUEST['URL2']
            REQUEST.RESPONSE.redirect(u+'/manage_main')
        return ''

    manage=manage_main=manage_editDocument=_redirect_up
